import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("ashape", "achunks", "ablocks"),
    {
        ((12, 10), (7, 5), (3, 3)),
        ((10,), (9,), (7,)),
    },
)
@pytest.mark.parametrize(
    ("bshape", "bchunks", "bblocks"),
    {
        ((10,), (4,), (2,)),
        ((10, 5), (3, 4), (1, 3)),
        ((10, 12), (2, 4), (1, 2)),
    },
)
@pytest.mark.parametrize(
    "dtype",
    {np.float32, np.float64},
)
def test_matmul(ashape, achunks, ablocks, bshape, bchunks, bblocks, dtype):
    a = blosc2.linspace(0, 1, dtype=dtype, shape=ashape, chunks=achunks, blocks=ablocks)
    b = blosc2.linspace(0, 1, dtype=dtype, shape=bshape, chunks=bchunks, blocks=bblocks)
    c = blosc2.matmul(a, b)

    na = a[:]
    nb = b[:]
    nc = np.matmul(na, nb)

    np.testing.assert_allclose(c, nc, rtol=1e-6)


@pytest.mark.parametrize(
    ("ashape", "achunks", "ablocks"),
    {
        ((12, 10), (7, 5), (3, 3)),
        ((10,), (9,), (7,)),
    },
)
@pytest.mark.parametrize(
    ("bshape", "bchunks", "bblocks"),
    {
        ((10,), (4,), (2,)),
        ((10, 5), (3, 4), (1, 3)),
        ((10, 12), (2, 4), (1, 2)),
    },
)
@pytest.mark.parametrize(
    "dtype",
    {np.complex64, np.complex128},
)
def test_complex(ashape, achunks, ablocks, bshape, bchunks, bblocks, dtype):
    real_part = blosc2.linspace(0, 1, shape=ashape, chunks=achunks, blocks=ablocks, dtype=dtype)
    imag_part = blosc2.linspace(0, 1, shape=ashape, chunks=achunks, blocks=ablocks, dtype=dtype)
    complex_matrix_a = real_part + 1j * imag_part
    a = blosc2.asarray(complex_matrix_a)

    real_part = blosc2.linspace(1, 2, shape=bshape, chunks=bchunks, blocks=bblocks, dtype=dtype)
    imag_part = blosc2.linspace(1, 2, shape=bshape, chunks=bchunks, blocks=bblocks, dtype=dtype)
    complex_matrix_b = real_part + 1j * imag_part
    b = blosc2.asarray(complex_matrix_b)

    c = blosc2.matmul(a, b)

    na = a[:]
    nb = b[:]
    nc = np.matmul(na, nb)

    np.testing.assert_allclose(c, nc, rtol=1e-6)


@pytest.mark.parametrize(
    ("ashape", "achunks", "ablocks"),
    {
        ((12, 11), (7, 5), (3, 1)),
        ((0, 0), (0, 0), (0, 0)),
        ((10,), (4,), (2,)),
    },
)
@pytest.mark.parametrize(
    ("bshape", "bchunks", "bblocks"),
    {
        ((1, 5), (1, 4), (1, 3)),
        ((4, 6), (2, 4), (1, 3)),
        ((5,), (4,), (2,)),
    },
)
def test_shapes(ashape, achunks, ablocks, bshape, bchunks, bblocks):
    a = blosc2.linspace(0, 10, shape=ashape, chunks=achunks, blocks=ablocks)
    b = blosc2.linspace(0, 10, shape=bshape, chunks=bchunks, blocks=bblocks)

    with pytest.raises(ValueError):
        blosc2.matmul(a, b)

    with pytest.raises(ValueError):
        blosc2.matmul(b, a)


@pytest.mark.parametrize(
    "scalar",
    {
        5,  # int
        5.3,  # float
        1 + 2j,  # complex
        np.int8(5),  # NumPy int8
        np.int16(5),  # NumPy int16
        np.int32(5),  # NumPy int32
        np.int64(5),  # NumPy int64
        np.float32(5.3),  # NumPy float32
        np.float64(5.3),  # NumPy float64
        np.complex64(1 + 2j),  # NumPy complex64
        np.complex128(1 + 2j),  # NumPy complex128
    },
)
def test_scalars(scalar):
    vector = blosc2.asarray(np.array([1, 2, 3]))

    with pytest.raises(ValueError):
        blosc2.matmul(scalar, vector)

    with pytest.raises(ValueError):
        blosc2.matmul(vector, scalar)

    with pytest.raises(ValueError):
        blosc2.matmul(scalar, scalar)


@pytest.mark.parametrize(
    "ashape",
    [
        (12, 10, 10),
        (3, 3, 3),
    ],
)
@pytest.mark.parametrize(
    "bshape",
    [
        (10, 10, 10, 11),
        (3, 2),
    ],
)
def test_dims(ashape, bshape):
    a = blosc2.linspace(0, 10, shape=ashape)
    b = blosc2.linspace(0, 1, shape=bshape)

    with pytest.raises(ValueError):
        blosc2.matmul(a, b)

    with pytest.raises(ValueError):
        blosc2.matmul(b, a)


@pytest.mark.parametrize(
    ("ashape", "achunks", "ablocks", "adtype"),
    {
        ((7, 10), (7, 5), (3, 5), np.float32),
        ((10,), (9,), (7,), np.complex64),
    },
)
@pytest.mark.parametrize(
    ("bshape", "bchunks", "bblocks", "bdtype"),
    {
        ((10,), (4,), (2,), np.float64),
        ((10, 6), (9, 4), (2, 3), np.complex128),
        ((10, 12), (2, 4), (1, 2), np.complex128),
    },
)
def test_special_cases(ashape, achunks, ablocks, adtype, bshape, bchunks, bblocks, bdtype):
    a = blosc2.linspace(0, 10, dtype=adtype, shape=ashape, chunks=achunks, blocks=ablocks)
    b = blosc2.linspace(0, 10, dtype=bdtype, shape=bshape, chunks=bchunks, blocks=bblocks)
    c = blosc2.matmul(a, b)

    na = a[:]
    nb = b[:]
    nc = np.matmul(na, nb)

    np.testing.assert_allclose(c, nc, rtol=1e-6)


def test_disk():
    a = blosc2.linspace(0, 1, shape=(3, 4), urlpath="a_test.b2nd", mode="w")
    b = blosc2.linspace(0, 1, shape=(4, 2), urlpath="b_test.b2nd", mode="w")
    c = blosc2.matmul(a, b, urlpath="c_test.b2nd", mode="w")

    na = a[:]
    nb = b[:]
    nc = np.matmul(na, nb)

    np.testing.assert_allclose(c, nc, rtol=1e-6)

    blosc2.remove_urlpath("a_test.b2nd")
    blosc2.remove_urlpath("b_test.b2nd")
    blosc2.remove_urlpath("c_test.b2nd")



from itertools import permutations

import numpy as np
import pytest

import blosc2


@pytest.fixture(
    params=[
        np.float64,
        pytest.param(np.int32, marks=pytest.mark.heavy),
        pytest.param(np.int64, marks=pytest.mark.heavy),
        pytest.param(np.float32, marks=pytest.mark.heavy),
    ]
)
def dtype_fixture(request):
    return request.param


@pytest.fixture(
    params=[
        ((10,), (5,), None),
        ((31,), (14,), (9,)),
    ]
)
def shape_1d_chunks_blocks(request):
    return request.param


@pytest.fixture(
    params=[
        ((3, 3), (2, 2), (1, 1)),
        ((12, 11), (7, 5), (6, 2)),
        ((1, 5), (1, 4), (1, 3)),
        pytest.param(((51, 603), (22, 99), (13, 29)), marks=pytest.mark.heavy),
    ]
)
def shape_2d_chunks_blocks(request):
    return request.param


@pytest.fixture(
    params=[
        ((4, 5, 2), (3, 4, 2), (3, 2, 1)),
        ((12, 10, 10), (11, 9, 7), (9, 7, 3)),
        pytest.param(((37, 63, 55), (12, 5, 41), (10, 5, 11)), marks=pytest.mark.heavy),
    ]
)
def shape_3d_chunks_blocks(request):
    return request.param


@pytest.fixture(
    params=[
        ((3, 3, 5, 7), (2, 3, 2, 4), (1, 1, 1, 4)),
        ((4, 6, 5, 2), (3, 3, 4, 2), (3, 2, 2, 1)),
        pytest.param(((10, 10, 10, 11), (7, 8, 9, 11), (6, 7, 8, 5)), marks=pytest.mark.heavy),
    ]
)
def shape_4d_chunks_blocks(request):
    return request.param


@pytest.mark.parametrize(
    "scalar",
    {
        1,  # int
        5.1,  # float
        1 + 2j,  # complex
        np.int8(2),  # NumPy int8
        np.int16(3),  # NumPy int16
        np.int32(4),  # NumPy int32
        np.int64(5),  # NumPy int64
        np.float32(5.2),  # NumPy float32
        np.float64(5.3),  # NumPy float64
        np.complex64(0 + 3j),  # NumPy complex64
        np.complex128(2 - 4j),  # NumPy complex128
    },
)
def test_scalars(scalar):
    at = blosc2.permute_dims(scalar)
    nat = np.transpose(scalar)

    np.testing.assert_allclose(at, nat)


def test_1d_transpose(shape_1d_chunks_blocks, dtype_fixture):
    shape, chunks, blocks = shape_1d_chunks_blocks
    a = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype_fixture)
    at = blosc2.permute_dims(a)

    na = a[:]
    nat = np.transpose(na)

    np.testing.assert_allclose(at, nat)


@pytest.mark.parametrize(
    "axes",
    list(permutations([0, 1])),
)
def test_2d_transpose(shape_2d_chunks_blocks, dtype_fixture, axes):
    shape, chunks, blocks = shape_2d_chunks_blocks
    a = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype_fixture)
    at = blosc2.permute_dims(a, axes=axes)

    na = a[:]
    nat = np.transpose(na, axes=axes)

    np.testing.assert_allclose(at, nat)


@pytest.mark.parametrize(
    "axes",
    list(permutations([0, 1, 2])),
)
def test_3d_transpose(shape_3d_chunks_blocks, dtype_fixture, axes):
    shape, chunks, blocks = shape_3d_chunks_blocks
    a = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype_fixture)
    at = blosc2.permute_dims(a, axes=axes)

    na = a[:]
    nat = np.transpose(na, axes=axes)

    np.testing.assert_allclose(at, nat)


@pytest.mark.parametrize(
    "axes",
    list(permutations([0, 1, 2, 3])),
)
def test_4d_transpose(shape_4d_chunks_blocks, dtype_fixture, axes):
    shape, chunks, blocks = shape_4d_chunks_blocks
    a = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype_fixture)
    at = blosc2.permute_dims(a, axes=axes)

    na = a[:]
    nat = np.transpose(na, axes=axes)

    np.testing.assert_allclose(at, nat)


@pytest.mark.heavy
@pytest.mark.parametrize(
    "axes",
    list(permutations([0, 1, 2])),
)
@pytest.mark.parametrize(
    "dtype",
    {np.complex64, np.complex128},
)
def test_complex(shape_3d_chunks_blocks, dtype, axes):
    shape, chunks, blocks = shape_3d_chunks_blocks
    real_part = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype)
    imag_part = blosc2.linspace(1, 0, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype)
    complex_matrix = real_part + 3j * imag_part

    a = blosc2.asarray(complex_matrix)
    at = blosc2.permute_dims(a, axes=axes)

    na = a[:]
    nat = np.transpose(na, axes=axes)

    np.testing.assert_allclose(at, nat)


@pytest.mark.parametrize(
    "axes",
    [
        (0, 0, 1),  # repeated axis
        (0, -1, -1),  # repeated negative
        (0, 1),  # missing one axis
        (0, 1, 2, 3),  # one more axis
        (0, 1, 3),  # out-of-range index
        (0, -4, 1),
    ],
)
def test_invalid_axes_raises(shape_3d_chunks_blocks, axes):
    shape, chunks, blocks = shape_3d_chunks_blocks
    a = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks)

    with pytest.raises(ValueError, match="not a valid permutation"):
        blosc2.permute_dims(a, axes=axes)


@pytest.mark.parametrize(
    "shape",
    [
        (2, 3),
        (4, 5, 6),
        (2, 4, 8, 5),
    ],
)
def test_matrix_transpose(shape):
    arr = blosc2.linspace(0, 1, shape=shape)
    result = blosc2.matrix_transpose(arr)

    expected = np.swapaxes(arr[:], -2, -1)

    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize(
    "shape",
    [
        (10,),
        (4, 5, 6),
        (2, 3, 4, 5),
    ],
)
def test_T_raises(shape):
    arr = blosc2.linspace(0, 1, shape=shape)
    with pytest.raises(ValueError, match="only works for 2-dimensional"):
        _ = arr.T


def test_disk():
    a = blosc2.linspace(0, 1, shape=(3, 4), urlpath="a_test.b2nd", mode="w")
    c = blosc2.permute_dims(a, urlpath="c_test.b2nd", mode="w")

    na = a[:]
    nc = np.transpose(na)

    np.testing.assert_allclose(c, nc, rtol=1e-6)

    blosc2.remove_urlpath("a_test.b2nd")
    blosc2.remove_urlpath("c_test.b2nd")


def test_transpose(shape_2d_chunks_blocks, dtype_fixture):
    shape, chunks, blocks = shape_2d_chunks_blocks
    a = blosc2.linspace(0, 1, shape=shape, chunks=chunks, blocks=blocks, dtype=dtype_fixture)
    with pytest.warns(DeprecationWarning):
        at = blosc2.transpose(a)

    na = a[:]
    nat = np.transpose(na)

    np.testing.assert_allclose(at, nat)



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Examples of using the jit decorator with arbitrary NumPy functions.
# These functions are not optimized for performance, but they show how
# to use the jit decorator with NumPy functions.
# You can find benchmarks for this example in the bench/ndarray directory

import numpy as np

import blosc2

# Create some sample data
a = blosc2.linspace(0, 1, 10 * 100, dtype="float32", shape=(10, 100))
b = blosc2.linspace(1, 2, 10 * 100, dtype="float32", shape=(10, 100))
c = blosc2.linspace(-10, 10, 10, dtype="float32", shape=(100,))


# Example 1: Basic usage of the jit decorator with reduction
@blosc2.jit
def expr_jit(a, b, c):
    # This function computes a cumulative sum reduction along axis 0
    return np.cumsum(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=0)


# Call the function with the jit decorator
result = expr_jit(a, b, c)
print(f"Example 1 result[0, 0:10]: {result[0, 0:10]}")


# Example 2: Using the jit decorator with an out parameter for reduction
out = np.zeros(result.shape, dtype=np.int64)


@blosc2.jit
def expr_jit_out(a, b, c):
    return np.cumulative_prod(((a**3 + np.sin(a * 2)) < c) & (b > 0), axis=0, out=out, include_initial=False)


# Call the function with the jit decorator and out parameter
result = expr_jit_out(a, b, c)
print(f"Example 2 result[0, 0:10]: {result[0, 0:10]}")
print("Example 2 out[0, 0:10] array:", out[0, 0:10])  # the 'out' array should now contain the same result


# Example 3: Using the jit decorator with a combination of NumPy functions
@blosc2.jit
def expr_jit_diff(a, b, c):
    return np.diff((a**3 + np.cumsum(b * 2, axis=1) + c), axis=1)


# Call the function with the jit decorator and custom parameters
result = expr_jit_diff(a, b, c)
print(f"Example 3 result[0, 0:5]: {result[0, 0:5]}")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import re
import sys


def update_version(new_version):
    # Update version in pyproject.toml
    with open("pyproject.toml") as file:
        pyproject_content = file.read()
    pyproject_content = re.sub(r'version = ".*"', f'version = "{new_version}"', pyproject_content)
    with open("pyproject.toml", "w") as file:
        file.write(pyproject_content)

    # Update version in src/blosc2/version.py
    with open("src/blosc2/version.py") as file:
        version_content = file.read()
    version_content = re.sub(r'__version__ = ".*"', f'__version__ = "{new_version}"', version_content)
    with open("src/blosc2/version.py", "w") as file:
        file.write(version_content)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python update_version.py <new_version>")
        sys.exit(1)
    new_version = sys.argv[1]
    update_version(new_version)
    print(f"Version updated to {new_version}")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

# Example for opening and reading a C2Array (remote array)

from time import time

import numpy as np

import blosc2

urlbase = "https://cat2.cloud/demo"
root = "@public"

# Access the server
# urlpath = blosc2.URLPath(f'{root}/examples/ds-1d.b2nd', urlbase)
# urlpath = blosc2.URLPath(f'{root}/examples/sa-1M.b2nd', urlbase)
urlpath = blosc2.URLPath(f"{root}/examples/lung-jpeg2000_10x.b2nd", urlbase)
# urlpath = blosc2.URLPath(f'{root}/examples/uncompressed_lung-jpeg2000_10x.b2nd', urlbase)

# Open the remote array
t0 = time()
remote_array = blosc2.open(urlpath, mode="r")
size = np.prod(remote_array.shape) * remote_array.cparams.typesize
print(f"Time for opening data (HTTP): {time() - t0:.3f}s - file size: {size / 2**10:.2f} KB")

# Fetch a slice of the remote array as a numpy array
t0 = time()
a = remote_array[5:9]
print(f"Time for reading data (HTTP): {time() - t0:.3f}s - {a.nbytes / 2**10:.2f} KB")

# TODO: Fetch a slice of the remote array as a blosc2.NDArray



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# This source code is licensed under a BSD-style license (found in the
# LICENSE file in the root directory of this source tree)
#######################################################################

import numpy as np
import pytest

import blosc2


@pytest.mark.parametrize(
    ("shape", "dtype"),
    [
        ((100, 1230), np.float64),
        ((23, 34), np.float32),
        ((80, 51, 60), "f4"),
        ((13, 13), None),
    ],
)
def test_nans_simple(shape, dtype):
    a = blosc2.nans(shape, dtype=dtype)
    assert a.dtype == np.dtype(dtype) if dtype is not None else np.dtype(np.float64)

    b = np.full(shape=shape, fill_value=np.nan, dtype=a.dtype)
    np.testing.assert_allclose(a[...], b)


@pytest.mark.parametrize("asarray", [True, False])
@pytest.mark.parametrize("typesize", [1, 3, 255, 256, 257, 256 * 256])
@pytest.mark.parametrize("shape", [(1,), (3,), (10,), (2 * 10,)])
def test_large_typesize(shape, typesize, asarray):
    dtype = np.dtype([("f_001", "f8", (typesize,)), ("f_002", "f4", (typesize,))])
    a = np.full(shape, np.nan, dtype=dtype)
    if asarray:
        b = blosc2.asarray(a)
    else:
        # b = blosc2.nans(shape, dtype=dtype)  # TODO: this is not working; perhaps deprecate blosc2.nans()?
        b = blosc2.full(shape, np.nan, dtype=dtype)
    for field in dtype.fields:
        np.testing.assert_allclose(b[field][:], a[field], equal_nan=True)
